/*
 * grid_accel.h
 * 
 * Molecular Dynamics simulation: Grid accelerator. 
 * 
 * Copyright (c) 2005 by Wolfgang Wieser ] wwieser (a) gmx <*> de [ 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

namespace NUM
{

template<typename P,bool nonzero_origin>class GridAccelerator
{
	public:
		typedef LinkedList<P> PList;
		
	private:
		// Universe size and origin. 
		fp3 u_size,u_origin;
		
		// Offset and scale vectors for universe to grid transformation. 
		fp3 u2g_off,u2g_scale;
		
		// Particle array holding all particles [npart]. 
		int npart;
		P *part;
		
		// Grid head list. 
		PList *lists;
		// List for all those which are in no other list. 
		PList outside_list;
		
		// Index into lists[] array: 
		inline PList *_lists_idx(int x,int y,int z)
			{  return(&lists[(z*gsy+y)*gsx+x]);  }
		
		// Size of the grid. 
		int gsx,gsy,gsz;
		
		// Translate coordinate from universe to grid coos. 
		// No range check yet. 
		inline void _u2gc(const fp3 &r,int &x,int &y,int &z)
		{
			if(nonzero_origin)
			{
				x = int( (r[0]-u2g_off[0])*u2g_scale[0] );
				y = int( (r[1]-u2g_off[1])*u2g_scale[1] );
				z = int( (r[2]-u2g_off[2])*u2g_scale[2] );
			}
			else
			{
				x = int( r[0]*u2g_scale[0] );
				y = int( r[1]*u2g_scale[1] );
				z = int( r[2]*u2g_scale[2] );
			}
		}
		// Translate coordinate into grid list. 
		inline PList *_u2gl(const fp3 &r)
		{
			int x,y,z;
			_u2gc(r,x,y,z);
			if(x<0 || y<0 || z<0 || x>=gsx || y>=gsy || z>=gsz)
			{  return(&outside_list);  }
			return(_lists_idx(x,y,z));
		}
		inline PList *_u2gl(const fp3 &r,int &x,int &y,int &z)
		{
			_u2gc(r,x,y,z);
			if(x<0 || y<0 || z<0 || x>=gsx || y>=gsy || z>=gsz)
			{  return(&outside_list);  }
			return(_lists_idx(x,y,z));
		}
		
	public:
		GridAccelerator()
		{
			lists=NULL;
			npart=0;
			part=NULL;
		}
		~GridAccelerator()
		{
			if(lists)
			{  free(lists);  }
			delete[] part;
		}
		
		// Get number of particles. 
		inline int n() const
			{  return(npart);  }
		
		// Get i-th particle: 
		inline P &operator[](int i)
			{  return(part[i]);  }
		
		// Get universe size. 
		inline fp3 USize() const  {  return(u_size);  }
		// Get universe origin. 
		inline fp3 UOrigin() const  {  return(u_origin);  }
		
		// Get grid size. 
		inline int gsize_x() const  {  return(gsx);  }
		inline int gsize_y() const  {  return(gsy);  }
		inline int gsize_z() const  {  return(gsz);  }
		
		// Set the size and the origin of the universe in universe units. 
		// And, set the grid size, i.e the number of grid cells we want to use. 
		void SetCoSy(const fp3 &origin,const fp3 &size,
			int gx,int gy,int gz)
		{
			u_size=size;
			u_origin=origin;
			
			u2g_off=origin;
			u2g_scale[0]=gx/u_size[0];
			u2g_scale[1]=gy/u_size[1];
			u2g_scale[2]=gz/u_size[2];
			
			gsx=gx;
			gsy=gy;
			gsz=gz;
			if(gsx && gsy && gsz)
			{
				size_t size = gsx*gsy*gsz*sizeof(PList);
				lists=(PList*)malloc(size);
				int n=gsx*gsy*gsz;
				for(int i=0; i<n; i++)
				{  new(&lists[i]) PList();  }
			}
			
			printf("Grid size: %d %d %d; cell size: %g %g %g\n",
				gsx,gsy,gsz, 1/u2g_scale[0], 1/u2g_scale[1], 1/u2g_scale[2]);
		}
		
		// Allocate a number of particles. 
		void AllocParticles(int n)
		{
			Assert(!part);
			npart=n;
			part=new P[npart];
		}
		
		// Insert particle into the accelerator. 
		// Only use during initial setup. 
		PList *insert(P *p)
		{
			PList *l=_u2gl(p->r);
			l->append(p);
			return(l);
		}
		
		// Get the grid cell particle list for this particle. 
		inline PList *u2gl(const P *p)
			{  return(_u2gl(p->r));  }
		inline PList *u2gl(const P *p,int &x,int &y,int &z)
			{  return(_u2gl(p->r,x,y,z));  }
		inline PList *u2gl(const fp3 &r,int &x,int &y,int &z)
			{  return(_u2gl(r,x,y,z));  }
		
// Get grid list... (unclean)
inline PList *lists_idx(int x,int y,int z)
	{  return(_lists_idx(x,y,z));  }
// Get outside grid list. (unclean)
inline PList *list_outside()
	{  return(&outside_list);  }
		
		// Change the crid cell for passed particle. 
		inline void ChangeGridCell(P *p,PList *old_list,PList *new_list)
		{
			if(old_list!=new_list)
			{
				old_list->dequeue(p);
				new_list->append(p);
			}
		}
		
		void DumpGridStat()
		{
			for(int yi=0; yi<gsy; yi++)
			{
				for(int zi=0; zi<gsz; zi++)
				{
					for(int xi=0; xi<gsx; xi++)
					{
						printf(" %3d",_lists_idx(xi,yi,zi)->count());
					}
					printf("\n");
				}
				printf("\n");
			}
		}
};

}  // End of namespace NUM. 
